from typing import *

from base_aux.aux_text.m5_re2_attemps import *
from base_aux.aux_attr.m4_kits import *
from base_aux.base_lambdas.m1_lambda import *
from base_aux.versions.m2_version import *


# =====================================================================================================================
class PatFormat:
    FIND_NAMES__IN_PAT: str = r"\{([_a-zA-Z]\w*)?([^{}]*)\}"   # (key, key_formatter)  dont use indexes!

    @classmethod
    @property
    def SPLIT_STATIC__IN_PAT(cls) -> str:
        result = r"(?:" + re.sub(r"\((.*?)\)", r"(?:\1)", cls.FIND_NAMES__IN_PAT) + r")"
        return result


# =====================================================================================================================
class TextFormatted(
    NestCall_Other,
    # NestRepr__ClsName_SelfStr     # use manual
):
    """
    GOAL
    ----
    access to formated values by value names

    SPECIALLY CREATED FOR
    ---------------------
    part for Alert messages

    NOTE
    ----
    1/ formatters is not tested and could not working!
    """
    PAT_FORMAT: str = ""    # FORMAT PATTERN
    VALUES: AttrDumped        # values set

    RAISE_TYPES: bool = False   # todo: decide to deprecate!

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, pat_format: str, *args: Any, raise_types: bool = None, **kwargs: Any) -> None:
        if raise_types is not None:
            self.RAISE_TYPES = raise_types

        self.PAT_FORMAT = pat_format

        self.init__keys()
        self.sai__values_args_kwargs(*args, **kwargs)
        self.init__types()

    # -----------------------------------------------------------------------------------------------------------------
    def init__keys(self) -> None:
        result_dict = {}
        for index, pat_group in enumerate(ReAttemptsAll(PatFormat.FIND_NAMES__IN_PAT).findall(self.PAT_FORMAT)):
            key, formatting = pat_group
            if not key:
                key = f"_{index}"
            result_dict.update({key: None})

        self.VALUES = AttrAux_AnnotsAll().annots__append_with_values(**result_dict)

    def sai__values_args_kwargs(self, *args, **kwargs) -> None | NoReturn:
        AttrAux_AnnotsAll(self.VALUES).sai__by_args_kwargs(*args, **kwargs)
        self.types__apply_on_values()

    def init__types(self) -> None:
        annots_dict = self.VALUES.__annotations__
        values_dict = AttrAux_AnnotsAll(self.VALUES).dump_dict()
        for name, type_i in annots_dict.items():
            if (type_i == Any) and (name in values_dict) and (values_dict[name] is not None):
                annots_dict[name] = type(values_dict[name])

    # -----------------------------------------------------------------------------------------------------------------
    def types__apply_on_values(self) -> None | NoReturn:
        annots_dict = self.VALUES.__annotations__
        values_dict = AttrAux_AnnotsAll(self.VALUES).dump_dict()
        for name, type_i in annots_dict.items():
            if (type_i != Any) and (name in values_dict) and (values_dict[name] is not None):
                value = values_dict[name]
                try:
                    value = type_i(value)
                except Exception as exc:
                    if self.RAISE_TYPES:
                        raise exc

                AttrAux_AnnotsAll(self.VALUES).sai__by_args_kwargs(**{name: value})

    def types__check_on_values(self) -> bool:
        """
        GOAL
        ----
        if you want to validate actual values
        """
        raise NotImplementedError()

    # -----------------------------------------------------------------------------------------------------------------
    # def __getattr__(self, item: str): # NOTE: DONT USE ANY GSAI HERE!!!
    #     return self[item]

    def __getitem__(self, item: str | int) -> Any | NoReturn:
        return IterAux(self.VALUES).value__get(item)

    # def __setattr__(self, item: str, value: Any):
    #     self[item] = value

    def __setitem__(self, item: str | int, value: Any) -> None | NoReturn:
        self.sai__values_args_kwargs(**{item: value})

    # -----------------------------------------------------------------------------------------------------------------
    def __str__(self) -> str:
        self.types__apply_on_values()
        result = str(self.PAT_FORMAT)
        values = AttrAux_AnnotsAll(self.VALUES).dump_dict()
        group_index = 0
        while True:
            match = re.search(PatFormat.FIND_NAMES__IN_PAT, result)
            if not match:
                break

            name, formatter = match.groups()
            name = name or f"_{group_index}"
            name_orig = IterAux(values).item__get_original(name)
            value = values[name_orig]
            if value is None:
                value = ""

            # apply type formatter ------
            try:
                formatter_type = formatter[-1]
                if formatter_type in ["s", ]:
                    value = str(value)
                elif formatter_type in ["d", "n"]:
                    value = int(value)
                elif formatter_type in ["f", "F"]:
                    value = float(value)
            except:
                pass

            # apply formatter -----------
            value_formatter = "{" + formatter + "}"
            try:
                value = value_formatter.format(value)
            except:
                pass

            result = re.sub(PatFormat.FIND_NAMES__IN_PAT, value, result, count=1)

            group_index += 1
        return result

    def __repr__(self):
        values = AttrAux_AnnotsAll(self.VALUES).dump_dict()
        result = f"{self.__class__.__name__}({self})"
        result += f"kwargs={values}"
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def other(self, other: str) -> Any | NoReturn:
        """
        GOAL
        ----
        reverse - parse result string back (get values)
        """
        static_data = re.split(PatFormat.SPLIT_STATIC__IN_PAT, self.PAT_FORMAT)
        pat_values_fullmatch = r""
        for static_i in static_data:
            if pat_values_fullmatch:
                pat_values_fullmatch += r"(.*)"

            pat_values_fullmatch += re.escape(static_i)

        values_match = re.fullmatch(pat_values_fullmatch, other)
        if values_match:
            values = values_match.groups()
            # values = [value.strip() for value in values]  # DONT DO STRIP!!!
            self.sai__values_args_kwargs(*values)
        else:
            raise Exc__Incompatible_Data(f"{other=}, {self.PAT_FORMAT=}")

        print(self)


# =====================================================================================================================
if __name__ == "__main__":
    pass


# =====================================================================================================================



from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_values.m4_primitives import *


# =====================================================================================================================
class Test__ensure:
    # @classmethod
    # def setup_class(cls):
    #     pass
    #     cls.Victim = type("Victim", (ValueUnit,), {})
    # @classmethod
    # def teardown_class(cls):
    #     pass
    #
    # def setup_method(self, method):
    #     pass
    #
    # def teardown_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            # DEF --------------
            (None, TYPES.NONE),
            ((None, ), tuple),

            (0, int),
            (0,int),
            ((0,), tuple),

            ((), tuple),
            ([], list),
            ({}, dict),
            ({1:1}, dict),

            # CALLABLES --------------
            (LAMBDA_TRUE, TYPES.FUNCTION),
            (LAMBDA_NONE, TYPES.FUNCTION),
            (LAMBDA_EXC, TYPES.FUNCTION),

            (VALUES_CALLABLE.METH_INST, TYPES.METHOD),
            (VALUES_CALLABLE.METH_CLS, TYPES.FUNCTION),

            (ClsGen, ClsGen),
            (INST_GEN, ClsGen),
        ]
    )
    def test__ensure_class(self, source, _EXPECTED):
        func_link = TypeAux(source).get__class
        Lambda(func_link).check_expected__assert(_EXPECTED)


# =====================================================================================================================



import time
import pytest
from base_aux.base_lambdas.m1_lambda import *
from base_aux.aux_datetime.m3_uptime import *


# =====================================================================================================================
START_AMENDMENT = 0.0000001


@pytest.mark.parametrize(
    argnames="pause",
    argvalues=[
        0.1, 0.2, 0.3,
    ]
)
def test__uptime(pause: float):
    ethalon = time.time()
    victim = Uptime(cmp_accuracy_value=START_AMENDMENT)
    assert victim.time_started - ethalon <= START_AMENDMENT
    assert victim <= START_AMENDMENT

    time.sleep(pause)
    assert victim.get() > pause - START_AMENDMENT
    assert victim > pause


# =====================================================================================================================



import pytest
import asyncio
from typing import *


# =====================================================================================================================
class RuleChar(NamedTuple):
    NAME: str
    VALUE: str
    SPACE_BEFORE: bool
    SPACE_AFTER: bool


# ---------------------------------------------------------------------------------------------------------------------
class StrChainCreation:
    """
    GOAL
    ----
    just a str creation! by GA!

    EXAMPLE
    -------
    assert StrChainCreation().Hello._COMMA.World._EXCLAMATION() == "Hello, World!"
    """
    _RESULT_: str = ""
    _RULES_TRANSLATE: set[RuleChar] = {
        RuleChar("_COMMA", ",", False, True),
        RuleChar("_SEMICOLON", ";", False, True),
        RuleChar("_COLON", ":", False, True),
        RuleChar("_ELLIPSIS", "...", False, True),
        RuleChar("_DASH", "-", False, False),
        RuleChar("_HYPHEN", "-", True, True),
        RuleChar("_SLASH", "/", False, False),
        RuleChar("_BSLASH", "\\", False, False),

        RuleChar("_SHARP", "#", True, False),
        RuleChar("_HASH", "#", True, False),
        RuleChar("_GRID", "#", True, False),

        RuleChar("_T", "\t", False, False),
        RuleChar("_TAB", "\t", False, False),
        RuleChar("_N", "\n", False, False),
        RuleChar("_NEWLINE", "\n", False, False),

        RuleChar("_D", ".", False, True),
        RuleChar("_DOT", ".", False, True),
        RuleChar("_Q", "?", False, True),
        RuleChar("_QUESTION", "?", False, True),
        RuleChar("_E", "!", False, True),
        RuleChar("_EXCLAMATION", "!", False, True),

        # TODO: finish!
        # TODO: add additional pattern into name item for SPACES! like
        #   assert victim.a._SHARP.b() == "a #b"    # default rule!!!
        #   assert victim.a._SHARP__TT.b() == "a # b"
        #   assert victim.a._SHARP__TF.b() == "a #b"
        #   assert victim.a._SHARP__FF.b() == "a#b"
    }

    def get_rule_char(self, item: str) -> RuleChar | None:
        # todo: item is NAME/VALUE!
        pass

    def __getattr__(self, item: str) -> Self:
        if item in self._RULES_TRANSLATE:
            item = self._RULES_TRANSLATE[item]

        if not self._RESULT_ or item in [rulechar.VALUE for rulechar in self._RULES_TRANSLATE]:
            additional = f"{item}"
        else:
            additional = f" {item}"

        self._RESULT_ += additional

        return self

    def __call__(self) -> str:
        return self._RESULT_


# =====================================================================================================================



from typing import *
from base_aux.loggers.m1_print import *

from base_aux.base_types.m0_static_aux import *
from base_aux.base_types.m1_type_aux import TypeAux
from base_aux.aux_attr.m0_static import check_name__buildin, NAMES__SKIP_PARTS


# =====================================================================================================================
def _value_search_by_list(source: Any, search_list: list[Any]) -> Any | None:
    search_list = search_list or []
    for search_item in search_list:
        try:
            if search_item in source:
                return search_item
        except:
            pass

        try:
            if search_item == source:
                return search_item
        except:
            pass


# =====================================================================================================================
class ObjectInfo:
    """
    GOAL
    ----
    print info about object (properties+methods results)

    But why? if we can use debugger directly?
    Reason:
    1. to get and save standard text info,
    it useful to keep this info for future quick eye sight without exact condition like other OS or device/devlist/configuration
    2. in debugger we cant see result of methods!
    try to see for example information from platform module! it have only methods and no one in object tree in debugger!
    ```python
    import platform

    obj = platform
    print(platform.platform())
    pass    # place debug point here
    ```
    3. Useful if you wish to see info from remote SOURCE if connecting directly over ssh for example

    FEATURES
        "print all properties/methods results",
        "show exceptions on methods/properties",
        "skip names by full/part names and use only by partnames",
        "separated collections by groups",

    TODO:
        "add TIMEOUT (use start in thread!) for print! use timeout for GETATTR!!!",
        [
            "realise PRINT_DIFFS=CHANGE_state/COMPARE_objects (one from different states like thread before and after start)!",
            "this is about to save object STATE!",
            "add parameter show only diffs or show all",
            "add TESTS after this step!",
        ],
        "apply asyncio.run for coroutine?",
        "merge items Property/Meth? - cause it does not matter callable or not (just add type info block)",
        "add check__instance_of_user_class",

    :ivar MAX_ITER_ITEMS: 0 or None if not limited!
    """
    # AUX -------------------------------------------------------------------------------------------------------------
    SOURCE: Any = None
    STATE: ObjectState = None
    STATE_OLD: ObjectState | None = None

    NAMES_COUNT__ON_START: int = 0
    NAMES_COUNT__ON_FINISH: int = 0

    # SETTINGS --------------------------------------------------------------------------------------------------------
    MAX_LINE_LEN: int = 100
    MAX_ITER_ITEMS: int = 5

    SKIP__BUILD_IN: bool = True

    NAMES__USE_ONLY_PARTS: list[str] = []
    NAMES__SKIP_FULL: list[str] = []
    NAMES__SKIP_PARTS: list[str] = NAMES__SKIP_PARTS

    HIDE_NAMES__TOUCHED: bool = None
    HIDE_NAMES__BUILDIN: bool = None

    def __init__(
            self,
            source: Optional[Any] = None,

            /,
            max_line_len: Optional[int] = None,
            max_iter_items: Optional[int] = None,
            skip__build_in: Optional[bool] = None,

            hide_names__touched: bool = None,
            hide_names__buildin: bool = None,

            names__use_only_parts: Union[None, str, list[str]] = None,
            names__skip_full: Union[None, str, list[str]] = None,
            names__skip_parts: str | list[str] = None,
    ):
        if source is not None:
            self.SOURCE = source

        # SETTINGS ----------------------------------------------------------------------------------------------------
        # RAPAMS -----------------------
        if max_line_len is not None:
            self.MAX_LINE_LEN = max_line_len
        if max_iter_items is not None:
            self.MAX_ITER_ITEMS = max_iter_items
        if skip__build_in is not None:
            self.SKIP__BUILD_IN = skip__build_in

        if hide_names__touched is not None:
            self.HIDE_NAMES__TOUCHED = hide_names__touched
        if hide_names__buildin is not None:
            self.HIDE_NAMES__BUILDIN = hide_names__buildin

        # NAMES LISTS -----------------------
        if names__use_only_parts:
            if isinstance(names__use_only_parts, str):
                names__use_only_parts = [names__use_only_parts, ]
            self.NAMES__USE_ONLY_PARTS = names__use_only_parts
        if names__skip_full:
            if isinstance(names__skip_full, str):
                names__skip_full = [names__skip_full, ]
            self.NAMES__SKIP_FULL.extend(names__skip_full)
        if names__skip_parts:
            if isinstance(names__skip_parts, str):
                names__skip_parts = [names__skip_parts, ]
            self.NAMES__SKIP_PARTS = [*self.NAMES__SKIP_PARTS, *names__skip_parts]

        # WORK --------------------------------------------------------------------------------------------------------
        self.state_reload()

    # =================================================================================================================
    @property
    def SOURCE_NAMES_COUNT_WAS_CHANGED(self) -> bool:
        return self.NAMES_COUNT__ON_START != self.NAMES_COUNT__ON_FINISH

    # =================================================================================================================
    def state_clear(self) -> None:
        self.STATE_OLD = self.STATE
        self.STATE = ObjectState()

    def state_reload(self) -> None:
        self.state_clear()

        # WORK --------------------------------------
        self._print_line__group_separator(f"{self.HIDE_NAMES__TOUCHED=}")
        self.NAMES_COUNT__ON_START = len(dir(self.SOURCE))
        print(f"{self.NAMES_COUNT__ON_START=}")

        # print()
        # print(dir(self.SOURCE))
        # for dirname in dir(self.SOURCE):
        #     print(dirname)
        # print()

        for pos, name in enumerate(dir(self.SOURCE), start=1):
            if not self.HIDE_NAMES__TOUCHED:
                print(f"{pos:4d}:{name}")

            # SKIP ----------------------------------------------------------------------------------------------------
            if self.SKIP__BUILD_IN and check_name__buildin(name):
                self.STATE.SKIPPED_BUILDIN.append(name)
                continue

            if name in self.NAMES__SKIP_FULL:
                self.STATE.SKIPPED_FULLNAMES.append(name)
                continue

            if _value_search_by_list(source=name, search_list=self.NAMES__SKIP_PARTS):
                self.STATE.SKIPPED_PARTNAMES.append(name)
                continue

            # FILTER --------------------------------------------------------------------------------------------------
            if self.NAMES__USE_ONLY_PARTS:
                use_name = False
                for name_include_item in self.NAMES__USE_ONLY_PARTS:
                    if name_include_item.lower() in name.lower():
                        use_name = True
                        break
                if not use_name:
                    continue

            # PROPERTIES/METHODS + Exception---------------------------------------------------------------------------
            attr_is_method: bool = False
            try:
                value = getattr(self.SOURCE, name)
            except Exception as exc:
                self.STATE.PROPERTIES__EXC.update({name: exc})
                continue

            if callable(value):
                attr_is_method = True
                try:
                    value = value()
                except Exception as exc:
                    self.STATE.METHODS__EXC.update({name: exc})
                    continue

            # print(f"{name=}/{attr_obj=}/type={type(attr_obj)}/elementary={isinstance(attr_obj, TYPES.ELEMENTARY)}")

            # PLACE VALUE ---------------------------------------------------------------------------------------------
            if TypeAux(value).check__elementary_single():
                if attr_is_method:
                    self.STATE.METHODS__ELEMENTARY_SINGLE.update({name: value})
                else:
                    self.STATE.PROPERTIES__ELEMENTARY_SINGLE.update({name: value})

            elif TypeAux(value).check__elementary_collection():
                if attr_is_method:
                    self.STATE.METHODS__ELEMENTARY_COLLECTION.update({name: value})
                else:
                    self.STATE.PROPERTIES__ELEMENTARY_COLLECTION.update({name: value})

            else:
                if attr_is_method:
                    self.STATE.METHODS__OBJECTS.update({name: value})
                else:
                    self.STATE.PROPERTIES__OBJECTS.update({name: value})

        self.NAMES_COUNT__ON_FINISH = len(dir(self.SOURCE))
        print(f"{self.NAMES_COUNT__ON_FINISH=}")

        msg = f"{self.SOURCE_NAMES_COUNT_WAS_CHANGED=}//{self.NAMES_COUNT__ON_START}/{self.NAMES_COUNT__ON_FINISH}"
        if self.SOURCE_NAMES_COUNT_WAS_CHANGED:
            Warn(msg)
        else:
            Print(msg)

    # def state_diff_last(self):
    #     # NOTE: use attrsDUMP!!! instead! cause raised values may be moved into other group! (single/collection)
    #     self.state_reload()
    #     result = ObjectState()
    #     for (name_g1, values_g1), (name_g2, values_g2) in zip(self.STATE_OLD.items(), self.STATE.items()):
    #         for

    # =================================================================================================================
    def _print_line__group_separator(self, group_name: str) -> None:
        """
        GOAL MAIN - print!
        GOAL SECONDARY - return str - just for tests!!!
        """
        result = "-" * 10 + f"{group_name:-<90}"      # here is standard MAX_LINE_LEN
        print(result)
        # return result

    def _print_line__name_type_value(self, name: Optional[str] = None, type_replace: Optional[str] = None, value: Union[None, Any, ItemKeyValue] = None, intend: Optional[int] = None) -> None:
        # -------------------------------
        name = name or ""
        if isinstance(value, ItemKeyValue):
            name = ""
        block_name = f"{name}"

        # -------------------------------
        block_type = f"{value.__class__.__name__}"
        if isinstance(value, ItemKeyValue):
            block_type = f"{value.KEY.__class__.__name__}:{value.VALUE.__class__.__name__}"
        if type_replace is not None:
            block_type = type_replace

        # -------------------------------
        intend = intend or 0
        if isinstance(value, ItemKeyValue):
            intend = 1

        _block_intend = "\t" * intend

        # -------------------------------
        try:
            block_value = f"{value}"
        except Exception as exc:
            block_value = f"{exc!r}"

        if isinstance(value, ItemKeyValue):
            block_type = f"{value.KEY}:{value.VALUE}"
        elif TypeAux(value).check__exception():
            block_value = f"{value!r}"

        # -------------------------------
        result = f"{block_name:20}\t{block_type:12}:{_block_intend}{block_value}"

        if self.MAX_LINE_LEN and len(result) > self.MAX_LINE_LEN:
            result = result[:self.MAX_LINE_LEN - 3*2] + "..."

        # --------------------------------------------------------------------------------------
        print(result)

        # -------------------------------
        try:
            if name and str(value) != repr(value) and str(value) != str(block_value) and not TypeAux(value).check__exception():
                # additional print repr()
                self._print_line__name_type_value(name=None, type_replace="__repr()", value=repr(value))
        except Exception as exc:
            print(f"{exc!r}")
            pass

        # return result

    # =================================================================================================================
    def _print_block__header(self) -> None:
        # start printing ----------------------------------
        group_name = f"{self.__class__.__name__}.print"
        self._print_line__group_separator(group_name.upper())

        try:
            print(f"str(SOURCE)={str(self.SOURCE)}")
        except Exception as exc:
            print(f"str(SOURCE)={exc!r}")

        try:
            print(f"repr(SOURCE)={repr(self.SOURCE)}")
        except Exception as exc:
            print(f"repr(SOURCE)={exc!r}")

        try:
            print(f"type(SOURCE)={type(self.SOURCE)}")
        except Exception as exc:
            print(f"type(SOURCE)={exc!r}")

        try:
            mro = self.SOURCE.__class__.__mro__
        except:
            mro = self.SOURCE.__mro__

        mro = [cls.__name__ for cls in mro]

        print(f"mro(SOURCE)={mro}")

        # SETTINGS ----------------------------------------
        group_name = "SETTINGS"
        self._print_line__group_separator(group_name)

        print(f"{self.SKIP__BUILD_IN=}")

        print(f"{self.NAMES__USE_ONLY_PARTS=}")
        print(f"{self.NAMES__SKIP_FULL=}")
        print(f"{self.NAMES__SKIP_PARTS=}")

        print(f"{self.MAX_LINE_LEN=}")
        print(f"{self.MAX_ITER_ITEMS=}")

    def _print_block__name_value(self, name, value) -> None:
        # ALWAYS ---------------------------------------------------------------------------------
        self._print_line__name_type_value(name=name, value=value)
        try:
            if len(str(value)) <= self.MAX_LINE_LEN:
                return
        except Exception as exc:
            print(f"{exc!r}")
            return

        # COLLECTION -----------------------------------------------------------------------------
        if TypeAux(value).check__elementary_collection():
            # start some pretty style -------------------------------------
            if not isinstance(value, dict):
                _index = 0
                for item in value:
                    _index += 1
                    if self.MAX_ITER_ITEMS and _index > self.MAX_ITER_ITEMS:
                        self._print_line__name_type_value(name=None, type_replace="", value="...", intend=1)
                        break
                    self._print_line__name_type_value(name=None, value=item, intend=1)

            elif isinstance(value, dict):
                _index = 0
                for item_key, item_value in value.items():
                    _index += 1
                    if self.MAX_ITER_ITEMS and _index > self.MAX_ITER_ITEMS:
                        self._print_line__name_type_value(name=None, type_replace="", value="...", intend=1)
                        break
                    self._print_line__name_type_value(name=None, value=ItemKeyValue(item_key, item_value))

        # SINGLE/EXC/OBJECTS ---------------------------------------------------------------------
        if any([
            TypeAux(value).check__elementary_single(),
            TypeAux(value).check__exception(),
            TypeAux(value).check__instance(),
        ]):
            pass    # DONT USE RETURN HERE OR ELIF IN NEXT LINE!!!

    # =================================================================================================================
    def print(self) -> None:
        """print all params from object
        if callable - try to call it!
        """
        WRAPPER_MAIN_LINE = "="*min(90, self.MAX_LINE_LEN)  # here we need use less then

        print(WRAPPER_MAIN_LINE)
        self._print_block__header()

        for group_name, group_values in self.STATE.__getstate__().items():
            self._print_line__group_separator(group_name)

            if group_name == "SKIPPED_BUILDIN" and self.HIDE_NAMES__BUILDIN:
                print(f"{self.HIDE_NAMES__BUILDIN=}")
                continue

            if TypeAux(group_values).check__elementary_collection_not_dict():
                for pos, name in enumerate(group_values, start=1):
                    print(f"{pos:4d}:{name}")
            else:
                for name, value in group_values.items():
                    self._print_block__name_value(name, value)

        print(WRAPPER_MAIN_LINE)

    # =================================================================================================================
    @classmethod
    def print_diffs(cls, *state: ObjectState) -> None:
        pass
        # TODO: FINISH!


# =====================================================================================================================
if __name__ == "__main__":
    class Victim:
        A1 = 1

        def meth1(self):
            return 1


    ObjectInfo(Victim()).print()
    assert True


# =====================================================================================================================



from base_aux.base_lambdas.m1_lambda2_derivatives import *
from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_values.m4_primitives import *


# =====================================================================================================================
def test__raise():
    try:
        Lambda(LAMBDA_RAISE)()
    except:
        pass
    else:
        assert False

    assert Lambda(INST_EXCEPTION)() == INST_EXCEPTION
    assert isinstance(Lambda(Exception)(), Exception)


# =====================================================================================================================
# DERIVATIVES
@pytest.mark.parametrize(
    argnames="source, args, _EXPECTED",
    argvalues=[
        (1, (1,2,), (1, True, False, True, False)),
        (10, (1,2,), (10, True, False, True, False)),
        (LAMBDA_TRUE, (1,2,), (True, True, False, True, False)),
        (LAMBDA_RAISE, (1,2,), (Exception, Exception, Exception, False, True)),
        (INST_CALL_RAISE, (1,2,), (Exception, Exception, Exception, False, True)),
        (INST_BOOL_RAISE, (1,2,), (INST_BOOL_RAISE, Exception, Exception, True, False)),
    ]
)
def test__derivatives(source, args, _EXPECTED):
    # for Cls, Expected in zip(, _EXPECTED):    # tis good idea but we cant see directly exact line!

    Lambda(source, *args).check_expected__assert(_EXPECTED[0])
    Lambda_Bool(source, *args).check_expected__assert(_EXPECTED[1])
    Lambda_BoolReversed(source, *args).check_expected__assert(_EXPECTED[2])
    Lambda_TrySuccess(source, *args).check_expected__assert(_EXPECTED[3])
    Lambda_TryFail(source, *args).check_expected__assert(_EXPECTED[4])


# =====================================================================================================================
def test__LambdaSleep_Ok():
    pause = 0.5

    start_time = time.time()
    victim = Lambda_Sleep(sec=pause, source=11)
    assert time.time() - start_time < 0.1
    assert victim == 11     # execute on EQ
    assert time.time() - start_time > pause * 0.9


def test__LambdaSleep_Raise():
    pause = 0.5
    start_time = time.time()
    victim = Lambda_Sleep(sec=pause, source=LAMBDA_RAISE)
    assert time.time() - start_time < 0.1
    try:
        result = victim == 11     # execute on EQ
    except:
        pass
    else:
        assert False
    assert time.time() - start_time > pause * 0.9


# =====================================================================================================================



from pathlib import Path

from base_aux.path2_file.m4_fileattrs import *
from base_aux.aux_datetime.m2_datetime import *
from base_aux.testplans.devices_kit import DeviceKit
from base_aux.breeders.m2_table_inst import *


# =====================================================================================================================
class Base_Stand:
    NAME: str = "[DEF] STAND NAME"
    DESCRIPTION: str = "[DEF] STAND DESCRIPTION"
    SN: str = "[DEF] STAND SN"

    DIRPATH_RESULTS: Union[str, Path] = "RESULTS"

    DEV_LINES: DeviceKit

    # TCSc_LINE: dict[type, bool]   # TODO: use TableLine??? - NO! KEEP DICT! with value like USING! so we can use one
    TCSc_LINE: TableLine = TableLine()

    TIMESTAMP_START: DateTimeAux | None = None
    TIMESTAMP_STOP: DateTimeAux | None = None

    # =================================================================================================================
    def __init__(self) -> None:
        # PREPARE CLSs ========================================
        for tc_cls in self.TCSc_LINE:
            # init STAND -----------------------------------
            tc_cls.STAND = self

            # gen INSTS -----------------------------------
            tcs_insts = []
            for index in range(self.DEV_LINES.COUNT_COLUMNS):
                tc_i = tc_cls(index=index)
                tcs_insts.append(tc_i)
            tc_cls.TCSi_LINE = TableLine(*tcs_insts)    # TODO: move into TC_CLS

    # =================================================================================================================
    def stand__get_info__general(self) -> dict[str, Any]:
        result = {
            "STAND.NAME": self.NAME,
            "STAND.DESCRIPTION": self.DESCRIPTION,
            "STAND.SN": self.SN,

            "STAND.TIMESTAMP_START": str(self.TIMESTAMP_START),
            "STAND.TIMESTAMP_STOP": str(self.TIMESTAMP_STOP),
        }
        return result

    def stand__get_info__tcs(self) -> dict[str, Any]:
        """
        get info/structure about stand/TP
        """
        TP_TCS = []
        for tc_cls in self.TCSc_LINE:
            TP_TCS.append(tc_cls.tcc__get_info())

        result = {
            "TESTCASES": TP_TCS,
            # "TP_DUTS": [],      # TODO: decide how to use
            # [
            #     # [{DUT1}, {DUT2}, …]
            #     {
            #         DUT_ID: 1  # ??? 	# aux
            #         DUT_SKIP: False
            #     }
            # ]

            }
        return result

    def stand__get_info__general_tcsc(self) -> dict[str, Any]:
        """
        get info/structure about stand/TP
===================================
some example
===================================
{
  "STAND.NAME": "[ОТК] БП800",
  "STAND.DESCRIPTION": "[DEF] STAND DESCRIPTION",
  "STAND.SN": "[DEF] STAND SN",
  "STAND.TIMESTAMP_START": "None",
  "STAND.TIMESTAMP_STOP": "None",
  "TESTCASES": [
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc0,ExtOff,HvOff,PsOff]присутствие БП",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc0,ExtOff,HvOff,PsOff]тест Заземления",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc0,ExtOff,HvOff,PsOff]проверка значений параметров ВЫКЛ",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc0,ExtOff,HvOff,PsOn]проверка значений параметров ВКЛ",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc0,ExtOn,HvOff,PsOff]тест PmBus",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc0,ExtOn,HvOff,PsOff]\nпроверка значений параметров ВЫКЛ",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc0,ExtOn,HvOff,PsOn]\nпроверка значений параметров ВКЛ",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc220,ExtOff,HvOn,PsOff]\nпроверка значений параметров в ВЫКЛ состоянии",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc220,ExtOff,HvOn,PsOn]\nпроверка значений параметров во ВКЛ состоянии",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc220,ExtOff,HvOn,PsOn]тест КЗ\nшины ожидания (STANDBY)",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc220,ExtOff,HvOn,PsOn]тест КЗ\nшины главного питания (MAIN)",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc220,ExtOff,HvOn,PsOn]тест нагрузочный\nшины ожидания (STANDBY)",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc220,ExtOff,HvOn,PsOn]тест нагрузочный\nшины главного питания (MAIN)",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc150,ExtOff,HvOn,PsOn]\nграница включения - нижняя (OFF)",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc160,ExtOff,HvOn,PsOn]\nграница включения - нижняя (ON)",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc250,ExtOff,HvOn,PsOn]\nграница включения - верхняя (ON)",
      "TC_ASYNC": true,
      "TC_SKIP": null
    },
    {
      "TC_NAME": "",
      "TC_DESCRIPTION": "[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)",
      "TC_ASYNC": true,
      "TC_SKIP": null
    }
  ]
}
        """
        result = {
            **self.stand__get_info__general(),
            **self.stand__get_info__tcs(),
        }
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def stand__get_results(self) -> dict[str, Any]:
        """
        get all results for stand/TP
===================================
some example
===================================
{
  "STAND": {
    "STAND.NAME": "[ОТК] БП800",
    "STAND.DESCRIPTION": "[DEF] STAND DESCRIPTION",
    "STAND.SN": "[DEF] STAND SN",
    "STAND.TIMESTAMP_START": "None",
    "STAND.TIMESTAMP_STOP": "None"
  },
  "TCS": {
    "": {       # NOTE: here is wrong! all TCS have no name! so all get balnk"" value and all are rewrite each other! need # FIXME
      "0": {
        "TC_NAME": "",
        "TC_DESCRIPTION": "[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)",
        "TC_ASYNC": true,
        "TC_SKIP": null,
        "DEV_FOUND": false,
        "INDEX": 0,
        "SKIP": null,
        "NAME": "PTB",
        "DESCRIPTION": "PTB for PSU",
        "SN": "",
        "FW": "",
        "MODEL": "",
        "DUT_SN": "",
        "DUT_FW": "",
        "DUT_MODEL": "",
        "timestamp_start": null,
        "tc_active": false,
        "tc_result_startup": false,
        "tc_result": null,
        "tc_details": {},
        "result__teardown": false,
        "timestamp_stop": null,
        "log": "INDEX=0\nDUT_SN=None\nDUT_ADDRESS=Enum__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID\ntc_skip_dut=None\nTC_NAME=\nTC_GROUP=Atc260\nTC_DESCRIPTION=[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)\nTC_ASYNC=True\nTC_SKIP=None\nSETTINGS=====================\nINFO_STR__ADD_ATTRS===========\nATC_VOUT:260\nPTB_SET_EXTON:False\nPTB_SET_HVON:True\nPTB_SET_PSON:True\nPROGRESS=====================\nSTATE_ACTIVE__CLS=EnumAdj_ProcessStateActive.NONE\ntimestamp_start=None\ntimestamp_stop=None\nexc=None\n------------------------------------------------------------\nresult__startup=None\nresult=None\nresult__teardown=None\n------------------------------------------------------------\nDETAILS=====================\n"
      },
      "1": {
        "TC_NAME": "",
        "TC_DESCRIPTION": "[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)",
        "TC_ASYNC": true,
        "TC_SKIP": null,
        "DEV_FOUND": false,
        "INDEX": 1,
        "SKIP": null,
        "NAME": "PTB",
        "DESCRIPTION": "PTB for PSU",
        "SN": "",
        "FW": "",
        "MODEL": "",
        "DUT_SN": "",
        "DUT_FW": "",
        "DUT_MODEL": "",
        "timestamp_start": null,
        "tc_active": false,
        "tc_result_startup": false,
        "tc_result": null,
        "tc_details": {},
        "result__teardown": false,
        "timestamp_stop": null,
        "log": "INDEX=1\nDUT_SN=None\nDUT_ADDRESS=Enum__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID\ntc_skip_dut=None\nTC_NAME=\nTC_GROUP=Atc260\nTC_DESCRIPTION=[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)\nTC_ASYNC=True\nTC_SKIP=None\nSETTINGS=====================\nINFO_STR__ADD_ATTRS===========\nATC_VOUT:260\nPTB_SET_EXTON:False\nPTB_SET_HVON:True\nPTB_SET_PSON:True\nPROGRESS=====================\nSTATE_ACTIVE__CLS=EnumAdj_ProcessStateActive.NONE\ntimestamp_start=None\ntimestamp_stop=None\nexc=None\n------------------------------------------------------------\nresult__startup=None\nresult=None\nresult__teardown=None\n------------------------------------------------------------\nDETAILS=====================\n"
      },
      "2": {
        "TC_NAME": "",
        "TC_DESCRIPTION": "[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)",
        "TC_ASYNC": true,
        "TC_SKIP": null,
        "DEV_FOUND": false,
        "INDEX": 2,
        "SKIP": null,
        "NAME": "PTB",
        "DESCRIPTION": "PTB for PSU",
        "SN": "",
        "FW": "",
        "MODEL": "",
        "DUT_SN": "",
        "DUT_FW": "",
        "DUT_MODEL": "",
        "timestamp_start": null,
        "tc_active": false,
        "tc_result_startup": false,
        "tc_result": null,
        "tc_details": {},
        "result__teardown": false,
        "timestamp_stop": null,
        "log": "INDEX=2\nDUT_SN=None\nDUT_ADDRESS=Enum__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID\ntc_skip_dut=None\nTC_NAME=\nTC_GROUP=Atc260\nTC_DESCRIPTION=[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)\nTC_ASYNC=True\nTC_SKIP=None\nSETTINGS=====================\nINFO_STR__ADD_ATTRS===========\nATC_VOUT:260\nPTB_SET_EXTON:False\nPTB_SET_HVON:True\nPTB_SET_PSON:True\nPROGRESS=====================\nSTATE_ACTIVE__CLS=EnumAdj_ProcessStateActive.NONE\ntimestamp_start=None\ntimestamp_stop=None\nexc=None\n------------------------------------------------------------\nresult__startup=None\nresult=None\nresult__teardown=None\n------------------------------------------------------------\nDETAILS=====================\n"
      },
      "3": {
        "TC_NAME": "",
        "TC_DESCRIPTION": "[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)",
        "TC_ASYNC": true,
        "TC_SKIP": null,
        "DEV_FOUND": false,
        "INDEX": 3,
        "SKIP": null,
        "NAME": "PTB",
        "DESCRIPTION": "PTB for PSU",
        "SN": "",
        "FW": "",
        "MODEL": "",
        "DUT_SN": "",
        "DUT_FW": "",
        "DUT_MODEL": "",
        "timestamp_start": null,
        "tc_active": false,
        "tc_result_startup": false,
        "tc_result": null,
        "tc_details": {},
        "result__teardown": false,
        "timestamp_stop": null,
        "log": "INDEX=3\nDUT_SN=None\nDUT_ADDRESS=Enum__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID\ntc_skip_dut=None\nTC_NAME=\nTC_GROUP=Atc260\nTC_DESCRIPTION=[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)\nTC_ASYNC=True\nTC_SKIP=None\nSETTINGS=====================\nINFO_STR__ADD_ATTRS===========\nATC_VOUT:260\nPTB_SET_EXTON:False\nPTB_SET_HVON:True\nPTB_SET_PSON:True\nPROGRESS=====================\nSTATE_ACTIVE__CLS=EnumAdj_ProcessStateActive.NONE\ntimestamp_start=None\ntimestamp_stop=None\nexc=None\n------------------------------------------------------------\nresult__startup=None\nresult=None\nresult__teardown=None\n------------------------------------------------------------\nDETAILS=====================\n"
      },
      "4": {
        "TC_NAME": "",
        "TC_DESCRIPTION": "[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)",
        "TC_ASYNC": true,
        "TC_SKIP": null,
        "DEV_FOUND": false,
        "INDEX": 4,
        "SKIP": null,
        "NAME": "PTB",
        "DESCRIPTION": "PTB for PSU",
        "SN": "",
        "FW": "",
        "MODEL": "",
        "DUT_SN": "",
        "DUT_FW": "",
        "DUT_MODEL": "",
        "timestamp_start": null,
        "tc_active": false,
        "tc_result_startup": false,
        "tc_result": null,
        "tc_details": {},
        "result__teardown": false,
        "timestamp_stop": null,
        "log": "INDEX=4\nDUT_SN=None\nDUT_ADDRESS=Enum__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID\ntc_skip_dut=None\nTC_NAME=\nTC_GROUP=Atc260\nTC_DESCRIPTION=[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)\nTC_ASYNC=True\nTC_SKIP=None\nSETTINGS=====================\nINFO_STR__ADD_ATTRS===========\nATC_VOUT:260\nPTB_SET_EXTON:False\nPTB_SET_HVON:True\nPTB_SET_PSON:True\nPROGRESS=====================\nSTATE_ACTIVE__CLS=EnumAdj_ProcessStateActive.NONE\ntimestamp_start=None\ntimestamp_stop=None\nexc=None\n------------------------------------------------------------\nresult__startup=None\nresult=None\nresult__teardown=None\n------------------------------------------------------------\nDETAILS=====================\n"
      },
      "5": {
        "TC_NAME": "",
        "TC_DESCRIPTION": "[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)",
        "TC_ASYNC": true,
        "TC_SKIP": null,
        "DEV_FOUND": false,
        "INDEX": 5,
        "SKIP": null,
        "NAME": "PTB",
        "DESCRIPTION": "PTB for PSU",
        "SN": "",
        "FW": "",
        "MODEL": "",
        "DUT_SN": "",
        "DUT_FW": "",
        "DUT_MODEL": "",
        "timestamp_start": null,
        "tc_active": false,
        "tc_result_startup": false,
        "tc_result": null,
        "tc_details": {},
        "result__teardown": false,
        "timestamp_stop": null,
        "log": "INDEX=5\nDUT_SN=None\nDUT_ADDRESS=Enum__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID\ntc_skip_dut=None\nTC_NAME=\nTC_GROUP=Atc260\nTC_DESCRIPTION=[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)\nTC_ASYNC=True\nTC_SKIP=None\nSETTINGS=====================\nINFO_STR__ADD_ATTRS===========\nATC_VOUT:260\nPTB_SET_EXTON:False\nPTB_SET_HVON:True\nPTB_SET_PSON:True\nPROGRESS=====================\nSTATE_ACTIVE__CLS=EnumAdj_ProcessStateActive.NONE\ntimestamp_start=None\ntimestamp_stop=None\nexc=None\n------------------------------------------------------------\nresult__startup=None\nresult=None\nresult__teardown=None\n------------------------------------------------------------\nDETAILS=====================\n"
      },
      "6": {
        "TC_NAME": "",
        "TC_DESCRIPTION": "[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)",
        "TC_ASYNC": true,
        "TC_SKIP": null,
        "DEV_FOUND": false,
        "INDEX": 6,
        "SKIP": null,
        "NAME": "PTB",
        "DESCRIPTION": "PTB for PSU",
        "SN": "",
        "FW": "",
        "MODEL": "",
        "DUT_SN": "",
        "DUT_FW": "",
        "DUT_MODEL": "",
        "timestamp_start": null,
        "tc_active": false,
        "tc_result_startup": false,
        "tc_result": null,
        "tc_details": {},
        "result__teardown": false,
        "timestamp_stop": null,
        "log": "INDEX=6\nDUT_SN=None\nDUT_ADDRESS=Enum__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID\ntc_skip_dut=None\nTC_NAME=\nTC_GROUP=Atc260\nTC_DESCRIPTION=[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)\nTC_ASYNC=True\nTC_SKIP=None\nSETTINGS=====================\nINFO_STR__ADD_ATTRS===========\nATC_VOUT:260\nPTB_SET_EXTON:False\nPTB_SET_HVON:True\nPTB_SET_PSON:True\nPROGRESS=====================\nSTATE_ACTIVE__CLS=EnumAdj_ProcessStateActive.NONE\ntimestamp_start=None\ntimestamp_stop=None\nexc=None\n------------------------------------------------------------\nresult__startup=None\nresult=None\nresult__teardown=None\n------------------------------------------------------------\nDETAILS=====================\n"
      },
      "7": {
        "TC_NAME": "",
        "TC_DESCRIPTION": "[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)",
        "TC_ASYNC": true,
        "TC_SKIP": null,
        "DEV_FOUND": false,
        "INDEX": 7,
        "SKIP": null,
        "NAME": "PTB",
        "DESCRIPTION": "PTB for PSU",
        "SN": "",
        "FW": "",
        "MODEL": "",
        "DUT_SN": "",
        "DUT_FW": "",
        "DUT_MODEL": "",
        "timestamp_start": null,
        "tc_active": false,
        "tc_result_startup": false,
        "tc_result": null,
        "tc_details": {},
        "result__teardown": false,
        "timestamp_stop": null,
        "log": "INDEX=7\nDUT_SN=None\nDUT_ADDRESS=Enum__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID\ntc_skip_dut=None\nTC_NAME=\nTC_GROUP=Atc260\nTC_DESCRIPTION=[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)\nTC_ASYNC=True\nTC_SKIP=None\nSETTINGS=====================\nINFO_STR__ADD_ATTRS===========\nATC_VOUT:260\nPTB_SET_EXTON:False\nPTB_SET_HVON:True\nPTB_SET_PSON:True\nPROGRESS=====================\nSTATE_ACTIVE__CLS=EnumAdj_ProcessStateActive.NONE\ntimestamp_start=None\ntimestamp_stop=None\nexc=None\n------------------------------------------------------------\nresult__startup=None\nresult=None\nresult__teardown=None\n------------------------------------------------------------\nDETAILS=====================\n"
      },
      "8": {
        "TC_NAME": "",
        "TC_DESCRIPTION": "[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)",
        "TC_ASYNC": true,
        "TC_SKIP": null,
        "DEV_FOUND": false,
        "INDEX": 8,
        "SKIP": null,
        "NAME": "PTB",
        "DESCRIPTION": "PTB for PSU",
        "SN": "",
        "FW": "",
        "MODEL": "",
        "DUT_SN": "",
        "DUT_FW": "",
        "DUT_MODEL": "",
        "timestamp_start": null,
        "tc_active": false,
        "tc_result_startup": false,
        "tc_result": null,
        "tc_details": {},
        "result__teardown": false,
        "timestamp_stop": null,
        "log": "INDEX=8\nDUT_SN=None\nDUT_ADDRESS=Enum__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID\ntc_skip_dut=None\nTC_NAME=\nTC_GROUP=Atc260\nTC_DESCRIPTION=[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)\nTC_ASYNC=True\nTC_SKIP=None\nSETTINGS=====================\nINFO_STR__ADD_ATTRS===========\nATC_VOUT:260\nPTB_SET_EXTON:False\nPTB_SET_HVON:True\nPTB_SET_PSON:True\nPROGRESS=====================\nSTATE_ACTIVE__CLS=EnumAdj_ProcessStateActive.NONE\ntimestamp_start=None\ntimestamp_stop=None\nexc=None\n------------------------------------------------------------\nresult__startup=None\nresult=None\nresult__teardown=None\n------------------------------------------------------------\nDETAILS=====================\n"
      },
      "9": {
        "TC_NAME": "",
        "TC_DESCRIPTION": "[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)",
        "TC_ASYNC": true,
        "TC_SKIP": null,
        "DEV_FOUND": false,
        "INDEX": 9,
        "SKIP": null,
        "NAME": "PTB",
        "DESCRIPTION": "PTB for PSU",
        "SN": "",
        "FW": "",
        "MODEL": "",
        "DUT_SN": "",
        "DUT_FW": "",
        "DUT_MODEL": "",
        "timestamp_start": null,
        "tc_active": false,
        "tc_result_startup": false,
        "tc_result": null,
        "tc_details": {},
        "result__teardown": false,
        "timestamp_stop": null,
        "log": "INDEX=9\nDUT_SN=None\nDUT_ADDRESS=Enum__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID\ntc_skip_dut=None\nTC_NAME=\nTC_GROUP=Atc260\nTC_DESCRIPTION=[Atc260,ExtOff,HvOn,PsOn]\nграница включения - верхняя (OFF)\nTC_ASYNC=True\nTC_SKIP=None\nSETTINGS=====================\nINFO_STR__ADD_ATTRS===========\nATC_VOUT:260\nPTB_SET_EXTON:False\nPTB_SET_HVON:True\nPTB_SET_PSON:True\nPROGRESS=====================\nSTATE_ACTIVE__CLS=EnumAdj_ProcessStateActive.NONE\ntimestamp_start=None\ntimestamp_stop=None\nexc=None\n------------------------------------------------------------\nresult__startup=None\nresult=None\nresult__teardown=None\n------------------------------------------------------------\nDETAILS=====================\n"
      }
    }
  }
}
        """
        TCS_RESULTS = {}
        for tc_cls in self.TCSc_LINE:
            TCS_RESULTS.update({tc_cls.NAME: tc_cls.tcsi__get_results()})

        result = {
            "STAND" : self.stand__get_info__general(),
            "TCS": TCS_RESULTS,
        }
        return result

    def stand__save_results(self) -> None:
        for index in range(self.DEV_LINES.COUNT_COLUMNS):
            result_i_short = {}
            result_i_full = {}
            for tc_cls in self.TCSc_LINE:
                tc_inst = None
                try:
                    tc_inst: 'Base_TestCase' = tc_cls.TCSi_LINE[index]

                    tc_inst_result_full = tc_inst.tci__get_result(add_info_dut=False, add_info_tc=False)
                    tc_inst_result_short = tc_inst_result_full["tc_result"]
                except:
                    tc_inst_result_short = None
                    tc_inst_result_full = None

                result_i_short.update({tc_cls.DESCRIPTION: tc_inst_result_short})
                result_i_full.update({tc_cls.DESCRIPTION: tc_inst_result_full})

            DUT = tc_inst.DEV_COLUMN.DUT

            if not DUT.DEV_FOUND or not DUT.DUT_FW:
                continue

            dut_info = DUT.dev__get_info()
            result_dut = {
                "STAND": self.stand__get_info__general(),
                "DUT": dut_info,
                "RESULTS_SHORT": result_i_short,
                "RESULTS_FULL": result_i_full,
            }

            # data_text = json.dumps(result_dut, indent=4, ensure_ascii=False)

            filename = f"{self.TIMESTAMP_STOP}[{index}].json"
            filepath = pathlib.Path(self.DIRPATH_RESULTS, filename)

            tfile = TextFile(text=str(result_dut), filepath=filepath)
            tfile.pretty__json()
            tfile.write__text()


# =====================================================================================================================
# if __name__ == "__main__":
#     print(load__tcs())


# =====================================================================================================================



from base_aux.base_lambdas.m1_lambda import *
from base_aux.aux_eq.m2_eq_aux import *

# =====================================================================================================================
class NestEq_AttrsNotPrivate:
    """
    GOAL
    ----
    mainly used for cmp bare attr-kits with no callables!

    LOGIC
    -----
    cmp first - direct callables
    cmp second - resolveExc!!!
    """
    # TODO: add __str/repr__
    def __eq__(self, other: Any) -> bool:
        # if isinstance() NestInit_AnnotsAttr_ByArgsKwargs == NestInit_AnnotsAttr_ByArgsKwargs:
        #     # check by names

        if other is None:
            return False

        try:
            for attr in AttrAux_Existed(self).iter__names_filter__not_private():
                # 1=cmp direct --------
                value_self_direct = getattr(self, attr)
                value_other_direct = getattr(other, attr)
                if EqAux(value_self_direct).check_doubleside__bool(value_other_direct):
                    continue

                # 2=cmp callables --------      # TODO: use EnumAdj_CallResolveStyle.SKIPCALLABLES ???
                value_self = Lambda(getattr, self, attr).resolve__exc()
                value_other = Lambda(getattr, other, attr).resolve__exc()

                if not EqAux(value_self).check_doubleside__bool(value_other):
                    return False

            return True
        except:
            return False


# =====================================================================================================================
class NestEq_AttrsNotHidden:
    def __eq__(self, other: Any) -> bool:
        # if isinstance() NestInit_AnnotsAttr_ByArgsKwargs == NestInit_AnnotsAttr_ByArgsKwargs:
        #     # check by names

        if other is None:
            return False

        try:
            for attr in AttrAux_Existed(self).iter__names_filter__not_hidden():
                # 1=cmp direct --------
                value_self_direct = getattr(self, attr)
                value_other_direct = getattr(other, attr)
                if EqAux(value_self_direct).check_doubleside__bool(value_other_direct):
                    continue

                # 2=cmp callables --------      # TODO: use EnumAdj_CallResolveStyle.SKIPCALLABLES ???
                value_self = Lambda(getattr, self, attr).resolve__exc()
                value_other = Lambda(getattr, other, attr).resolve__exc()

                if not EqAux(value_self).check_doubleside__bool(value_other):
                    return False

            return True
        except:
            return False


# =====================================================================================================================



import threading

import MetaTrader5 as mt5

from base_aux.aux_np_pd.m0_typing  import *
from base_aux.aux_np_pd.m2_time_series import *
from base_aux.aux_np_pd.m3_indicators import *
from base_aux.stock.m0_symbols import *

from base_aux.privates.m1_privates import *
from base_aux.alerts.m1_alert0_base import *
from base_aux.base_types.m2_info import *
from base_aux.aux_attr.m4_kits import *
from base_aux.loggers.m1_print import *


# =====================================================================================================================
# TODO: ADD STABILITY STRESSFUL! like in IMAP!!!! connect if lost!


# =====================================================================================================================
TYPING__SYMBOL_FINAL = mt5.SymbolInfo
TYPING__SYMBOL_DRAFT = Union[str, mt5.SymbolInfo]
TYPING__TF = int
TYPING__INDICATOR_VALUES = Union[None, float, TYPING__PD_SERIES]


# =====================================================================================================================
class MT5(NestInit_AttrsLambdaResolve):
    """
    GOAL
    ----
    MAIN
        1/ connect to mt5   # TODO: move mt5 into classAttr! to use connection for one time only
        2/ get history
    EXTRA
        3/ do smth universal things like getting Available symbols

    DONT get/calculate indicators!!!
    """
    CONN_AUTH: Base_AttrKit = PvLoaderIni_AuthServer(keypath=("AUTH_MT5_DEF",))
    SYMBOL: TYPING__SYMBOL_FINAL = Symbols.BRENT_UNIVERSAL
    TF: TYPING__TF = mt5.TIMEFRAME_D1
    __SYMBOLS_AVAILABLE: list[mt5.SymbolInfo] = None

    # BAR_LAST: np.ndarray = None
    """
    time
    (1675118400, 85.41, 85.43, 85.21, 85.21, 225, 1, 1065)
    bar["time"] --> 1675118400
    """
    # TICK_LAST: mt5.Tick = None
    """
    self.BAR_LAST_TICK=Tick(time=1675468684, bid=83.45, ask=83.51, last=83.5, volume=6, time_msc=1675468684950, flags=30, volume_real=6.0)
    type(self.BAR_LAST_TICK)=<class 'Tick'>
    """

    _symbols__volume_price: dict[str, float] = {}     # collect in threads! dont delete!

    # =================================================================================================================
    def __init__(
            self,
            tf: TYPING__TF = None,
            symbol: TYPING__SYMBOL_DRAFT = None
    ) -> None | NoReturn:
        super().__init__()

        if tf is not None:
            self.TF = tf
        if symbol is not None:
            self.SYMBOL = symbol or self.SYMBOL

        self.mt5_connect()
        self._SYMBOL_init()

    def __del__(self):
        mt5.shutdown()

    # CONNECT ---------------------------------------------------------------------------------------------------------
    def mt5_connect(self) -> None | NoReturn:
        result = mt5.initialize(login=int(self.CONN_AUTH.NAME), password=self.CONN_AUTH.PWD, server=self.CONN_AUTH.SERVER)
        msg = f"[{result}]initialize[{mt5.last_error()=}]"
        Print(msg)
        if not result:
            msg += f"SMTIMES PWD DROPPED_DOWN/CORRUPTED in MT5!!! - you should simply update it in MT5"
            msg += f"\n{self.CONN_AUTH}"
            Warn(msg)
            raise ConnectionError(msg)

    # SYMBOL ==========================================================================================================
    def _SYMBOL_init(self) -> None | NoReturn:
        self.SYMBOL = self.SYMBOL__get_active()

    def SYMBOL__get_active(self, _symbol: TYPING__SYMBOL_DRAFT = None) -> Union[mt5.SymbolInfo, NoReturn]:
        _symbol = _symbol or self.SYMBOL
        if isinstance(_symbol, str):
            _symbol = mt5.symbol_info(_symbol)
            last_error = mt5.last_error()
            if last_error[0] != 1:
                msg = f"incorrect {_symbol=}/{last_error=}"
                raise Exc__NotExistsNotFoundNotCreated(msg)

        if not isinstance(_symbol, mt5.SymbolInfo):
            msg = f"incorrect {_symbol=}"
            raise Exc__NotExistsNotFoundNotCreated(msg)

        return _symbol

    def TF__get_active(self, _tf: TYPING__TF = None) -> TYPING__TF:
        return _tf or self.TF

    # AVAILABLE -------------------------------------------------------------------------------------------------------
    @classmethod
    @property
    def SYMBOLS_AVAILABLE(cls) -> list[mt5.SymbolInfo]:
        """
        too much time to get all items! dont use it without special needs!
        """
        if not cls.__SYMBOLS_AVAILABLE:
            cls.__SYMBOLS_AVAILABLE = mt5.symbols_get()
        return cls.__SYMBOLS_AVAILABLE

    def symbols_available__by_mask(
            self,
            mask: str = "",
            only_rus: bool = False
    ) -> list[mt5.SymbolInfo]:
        # count=12976 ALL!!!!
        # count=275 rus!!!!
        count = 0
        result = []

        if not mask:
            symbols = self.SYMBOLS_AVAILABLE
        else:
            symbols = list(filter(lambda x: x.name.startswith(mask), self.SYMBOLS_AVAILABLE))

        print("*"*100)
        for item in symbols:
            # FILTER LONG NAMES like for Options
            if len(item.name) > 5:
                continue

            # FILTER RUS
            if only_rus:
                if any([
                    item.isin != "moex.stock",
                    not re.match(pattern=r"[а-яА-Я]", string=item.description),
                    item.name.startswith("RU00"),
                    # item.currency_base != "RUS",  # always OK!
                    # "-RM" in item.name,     #count=12301
                ]):
                    continue

            # GET
            count += 1
            result.append(item)
            print(item.name)

        print("*"*100)
        print(f"result={[item.name for item in result]}")
        print(f"{count=}")
        print("*"*100)

        return result

    # SHOW ------------------------------------------------------------------------------------------------------------
    def _mt5_symbol_show(self, show: bool = True, _symbol: TYPING__SYMBOL_DRAFT = None) -> bool:
        _symbol = self.SYMBOL__get_active(_symbol)
        result = mt5.symbol_select(_symbol.name, show)
        print(f"[{result}]_mt5_symbol_show({_symbol})={mt5.last_error()=}")
        return result

    def _mt5_symbol_show__check(self, _symbol: TYPING__SYMBOL_DRAFT = None) -> bool:
        _symbol = self.SYMBOL__get_active(_symbol)
        if _symbol:
            return _symbol.select

    # INFO ------------------------------------------------------------------------------------------------------------
    def _symbols_info__print_compare(self, symbols: list[TYPING__SYMBOL_DRAFT] = ["SBER", "AAPL-RM", "PYPL-RM"]):
        """
        since SYMBOL_NAME not added into chart gui list - it will return zero to many attributes!

        ****************************************************************************************************
        INSTRUMENTS                   =['SBER', 'AAPL-RM', 'PYPL-RM']
        ****************************************************************************************************
        custom                        =[False, False, False]
        chart_mode                    =[1, 1, 1]
        select                        =[True, False, False]
        visible                       =[True, False, False]
        session_deals                 =[0, 0, 0]
        session_buy_orders            =[0, 0, 0]
        session_sell_orders           =[0, 0, 0]
        volume                        =[34, 0, 0]
        volumehigh                    =[12811, 0, 0]
        volumelow                     =[1, 0, 0]
        time                          =[1671839399, 0, 0]
        digits                        =[2, 0, 0]
        spread                        =[6, 0, 0]
        spread_float                  =[True, True, True]
        ticks_bookdepth               =[32, 32, 32]
        trade_calc_mode               =[32, 32, 32]
        trade_mode                    =[4, 4, 4]
        start_time                    =[0, 0, 0]
        expiration_time               =[0, 0, 0]
        trade_stops_level             =[0, 0, 0]
        trade_freeze_level            =[0, 0, 0]
        trade_exemode                 =[3, 3, 3]
        swap_mode                     =[0, 0, 0]
        swap_rollover3days            =[3, 3, 3]
        margin_hedged_use_leg         =[False, False, False]
        expiration_mode               =[15, 15, 15]
        filling_mode                  =[3, 3, 3]
        order_mode                    =[63, 63, 63]
        order_gtc_mode                =[2, 2, 2]
        option_mode                   =[0, 0, 0]
        option_right                  =[0, 0, 0]
        bid                           =[137.85, 0.0, 0.0]
        bidhigh                       =[139.01, 0.0, 0.0]
        bidlow                        =[136.81, 0.0, 0.0]
        ask                           =[137.91, 0.0, 0.0]
        askhigh                       =[138.36, 0.0, 0.0]
        asklow                        =[136.82, 0.0, 0.0]
        last                          =[137.94, 0.0, 0.0]
        lasthigh                      =[138.26, 0.0, 0.0]
        lastlow                       =[136.81, 0.0, 0.0]
        volume_real                   =[34.0, 0.0, 0.0]
        volumehigh_real               =[12811.0, 0.0, 0.0]
        volumelow_real                =[1.0, 0.0, 0.0]
        option_strike                 =[0.0, 0.0, 0.0]
        point                         =[0.01, 1.0, 1.0]
        trade_tick_value              =[0.1, 1.0, 1.0]
        trade_tick_value_profit       =[0.1, 1.0, 1.0]
        trade_tick_value_loss         =[0.1, 1.0, 1.0]
        trade_tick_size               =[0.01, 1.0, 1.0]
        trade_contract_size           =[10.0, 1.0, 1.0]
        trade_accrued_interest        =[0.0, 0.0, 0.0]
        trade_face_value              =[0.0, 0.0, 0.0]
        trade_liquidity_rate          =[1.0, 1.0, 1.0]
        volume_min                    =[1.0, 1.0, 1.0]
        volume_max                    =[100000000.0, 100000000.0, 100000000.0]
        volume_step                   =[1.0, 1.0, 1.0]
        volume_limit                  =[0.0, 0.0, 0.0]
        swap_long                     =[0.0, 0.0, 0.0]
        swap_short                    =[0.0, 0.0, 0.0]
        margin_initial                =[0.0, 0.0, 0.0]
        margin_maintenance            =[0.0, 0.0, 0.0]
        session_volume                =[0.0, 0.0, 0.0]
        session_turnover              =[0.0, 0.0, 0.0]
        session_interest              =[0.0, 0.0, 0.0]
        session_buy_orders_volume     =[0.0, 0.0, 0.0]
        session_sell_orders_volume    =[0.0, 0.0, 0.0]
        session_open                  =[137.49, 0.0, 0.0]
        session_close                 =[137.69, 0.0, 0.0]
        session_aw                    =[0.0, 0.0, 0.0]
        session_price_settlement      =[0.0, 0.0, 0.0]
        session_price_limit_min       =[0.0, 0.0, 0.0]
        session_price_limit_max       =[0.0, 0.0, 0.0]
        margin_hedged                 =[0.0, 0.0, 0.0]
        price_change                  =[0.1816, 0.0, 0.0]
        price_volatility              =[0.0, 0.0, 0.0]
        price_theoretical             =[0.0, 0.0, 0.0]
        price_greeks_delta            =[0.0, 0.0, 0.0]
        price_greeks_theta            =[0.0, 0.0, 0.0]
        price_greeks_gamma            =[0.0, 0.0, 0.0]
        price_greeks_vega             =[0.0, 0.0, 0.0]
        price_greeks_rho              =[0.0, 0.0, 0.0]
        price_greeks_omega            =[0.0, 0.0, 0.0]
        price_sensitivity             =[0.0, 0.0, 0.0]
        basis                         =['', '', '']
        category                      =['', '', '']
        currency_base                 =['RUR', 'RUR', 'RUR']
        currency_profit               =['RUR', 'RUR', 'RUR']
        currency_margin               =['RUR', 'RUR', 'RUR']
        bank                          =['', '', '']
        description                   =['Сбербанк России ПАО ао', 'Apple Inc.', 'PayPal Holdings, Inc.']
        exchange                      =['', '', '']
        formula                       =['', '', '']
        isin                          =['moex.stock', 'moex.stock', 'moex.stock']
        name                          =['SBER', 'AAPL-RM', 'PYPL-RM']
        page                          =['', '', '']
        path                          =['MOEX\\SBER', 'MOEX\\AAPL-RM', 'MOEX\\PYPL-RM']
        ****************************************************************************************************
        """
        items = []
        for symbol in list(symbols):
            item = self.SYMBOL__get_active(symbol)
            if item:
                items.append(item._asdict())
            else:
                symbols.remove(symbol)

        if not items:
            return

        print("*"*100)
        key = "ACTIVES"
        print(f"{key:30}={symbols}")
        print("*"*100)
        for key in items[0]:
            value = []
            for item in items:
                value.append(item.get(key))
            print(f"{key:30}={value}")

        print("*"*100)

    # VOLUME_PRICE -----------------------------------------------------
    def _symbol__get_volume_price(self, _symbol: TYPING__SYMBOL_DRAFT = None, _devider: Optional[int] = None) -> float:
        """
        VolumePrice as priceMean * Volume
        +save result into self.symbols_volume_price for threading usage!

        value will be differ from official because i get mean price!
        https://www.moex.com/ru/marketdata/?g=4#/mode=groups&group=4&collection=3&boardgroup=57&data_type=current&category=main
        """
        _devider = _devider or 1000 * 1000
        bar = self.history__get(_symbol=_symbol, _tf=mt5.TIMEFRAME_D1)
        # print(f"{bar['real_volume']=}")

        item = mt5.symbol_info(_symbol)
        contracts_per_lot = item.trade_contract_size
        contracts = contracts_per_lot * bar["real_volume"]
        volume_price = contracts * (bar["high"] + bar["low"])/2
        # print(f"{volume_price=}")

        result = round(volume_price[0]/_devider)
        self._symbols__volume_price.update({_symbol: result})    # dont delete! collect in threads!
        return result

    def _symbols__get_volume_price__sorted(self, limit_min=None, limit_max=None, _symbols: Optional[list[str]] = None, _devider: Optional[int] = None) -> dict[str, float]:
        """

        (400 * 1000 * 1000)
        ['SBER', 'GAZP', 'LKOH', 'PLZL', 'MGNT', 'NVTK', 'UWGN', 'LQDT', 'VTBR', 'GMKN', 'LSNG', 'SNGS', 'ROSN', 'CHMF']

        :param limit_min:
        :param limit_max:

{
    "GAZP": 18552,
    "SBER": 16418,
    "LQDT": 14222,
    "TCSG": 10671,
    "VTBR": 4788,
    "FIVE": 4602,
    "LKOH": 4085,
    "NVTK": 4052,
    "GMKN": 3677,
    "PLZL": 3272,
    "AFKS": 2283,
    "ROSN": 2160,
    "TATN": 2115,
    "AFLT": 2069,
    "SELG": 1995,
    "MGNT": 1944,
    "MOEX": 1790,
    "OZON": 1661,
    "SMLT": 1563,
    "WUSH": 1341,
    "RUAL": 1313,
    "SBERP": 1304,
    "CHMF": 1266,
    "MTLR": 1199,
    "RNFT": 1162,
    "SNGS": 1143,
    "NLMK": 1092,
    "PIKK": 1079,
    "IRKT": 975,
    "MAGN": 970,
    "TRNFP": 936,
    "SPBE": 889,
    "UPRO": 882,
    "VKCO": 862,
    "ALRS": 788,
    "SNGSP": 788,
    "MTSS": 772,
    "RTKM": 746,
    "GRNT": 745,
    "ENPG": 704,
    "FLOT": 703,
    "SGZH": 697,
    "SIBN": 695,
    "UWGN": 639,
    "AGRO": 589,
    "TRUR": 565,
    "BELU": 557,
    "TATNP": 554,
    "PHOR": 503,
    "TMOS": 497,
    "IRAO": 485,
    "TGLD": 450,
    "SFIN": 445,
    "TRMK": 399,
    "EQMX": 388,
    "DASB": 333,
    "MVID": 330,
    "BSPB": 307,
    "MTLRP": 274,
    "GTRK": 268,
    "POSI": 263,
    "UNAC": 247,
    "HYDR": 235,
    "RENI": 224,
    "AKME": 204,
    "GLTR": 186,
    "RASP": 183,
    "BANEP": 178,
    "FESH": 177,
    "FEES": 158,
    "VSMO": 153,
    "HHRU": 144,
    "AQUA": 125,
    "ISKJ": 122,
    "RBCM": 122,
    "TGKN": 121,
    "SVAV": 118,
    "KMAZ": 111,
    "SBMX": 109,
    "MDMG": 107,
    "CBOM": 104
}
        """
        _devider = _devider or 1000 * 1000
        limit_min = limit_min if limit_min is not None else 100 * 1000 * 1000 / _devider
        _symbols = _symbols or Symbols.SYMBOLS__RUS_FINAM

        # LOAD ---------------------------------------------------
        for symbol in _symbols:
            threading.Thread(target=self._symbol__get_volume_price, kwargs=dict(_symbol=symbol, _devider=_devider)).start()

        while threading.active_count() > 1:
            time.sleep(1)

        # FILTER ---------------------------------------------------
        for symbol, value in dict(self._symbols__volume_price).items():
            if limit_max and limit_max < value:
                self._symbols__volume_price.pop(symbol)
            if limit_min > value:
                self._symbols__volume_price.pop(symbol)

        # SORT -----------------------------------------------------
        self._symbols__volume_price = dict(sorted(self._symbols__volume_price.items(), key=lambda x: x[1], reverse=True))

        # PRINT ----------------------------------------------------
        result_pretty = json.dumps(self._symbols__volume_price, indent=4)
        print(result_pretty)
        return self._symbols__volume_price

    # BAR HISTORY =====================================================================================================
    def history__get(
            self,
            count: int = 1,
            tf_multiply: int = None,
            _start: int = None,
            _symbol: TYPING__SYMBOL_DRAFT = None,
            _tf: TYPING__TF = None
    ) -> np.ndarray:
        """
        GOAL
        ----
        get history bars
        :param tf_multiply: correct count of bars in case of using increasing tf

        :_start: 0 is actual and not finished!
        ['time', 'open', 'high', 'low', 'close', 'tick_volume', 'spread', 'real_volume']
            [(1695763800, 93.83, 93.88, 93.78, 93.88, 172, 1, 723)]
            elem=1695763800/<class 'numpy.int64'>
            elem=93.83/<class 'numpy.float64'>
            elem=93.88/<class 'numpy.float64'>
            elem=93.78/<class 'numpy.float64'>
            elem=93.88/<class 'numpy.float64'>
            elem=172/<class 'numpy.uint64'>
            elem=1/<class 'numpy.intc'>
            elem=723/<class 'numpy.uint64'>

        returns
            1 bars
                [(1741999200, 70.62, 70.62, 70.62, 70.62, 10, 3, 10)]
                ndim                	int         :1
                size                	int         :1
            2 bars
                [(1741998600, 70.61, 70.62, 70.61, 70.62,  7, 3,  7)
                 (1741999200, 70.62, 70.62, 70.62, 70.62, 10, 3, 10)]
                ndim                	int         :1
                size                	int         :2
        """
        _symbol = self.SYMBOL__get_active(_symbol)
        _tf = self.TF__get_active(_tf)
        tf_multiply = tf_multiply or 1

        if _start is None:
            _start = 1

        bars = mt5.copy_rates_from_pos(_symbol.name, _tf, _start, count * tf_multiply)
        # if not bars:
        #     print(f"{_symbol=}/{bars=}")
        #     return

        # for bar in bars:
        #     print(f"{type(bar)}={bar}")     # <class 'numpy.void'>=(1671753600, 137.49, 138.26, 136.81, 137.94, 53823, 0, 2283422)

        if tf_multiply > 1:
            bars = NpTimeSeriesAux(bars).shrink(tf_multiply)

        # if count == 1:
        #     # bars = [(1695729000, 92.3, 92.42, 92.22, 92.23, 944, 1, 3381)]
        #     return bars[0]  # numpy.void
        # else:
        #     # bars = [(1695728400, 92.16, 92.32, 92.1, 92.31, 578, 1, 2764)
        #     #         (1695729000, 92.3, 92.42, 92.22, 92.23, 944, 1, 3381)]
        #     return bars     # numpy.ndarray

        return bars

    def history_new__wait(self, old: np.ndarray, sleep: int = 10) -> None:
        count = 0
        new = None
        while not new or old == new:
            count += 1
            try:
                new = self.history__get()[0]
                break
            except:
                pass

            print(f"bar_new__wait {count=}")
            time.sleep(sleep)

    # -----------------------------------------------------------------------------------------------------------------
    # def tick_last__update(self, _symbol: TYPING__SYMBOL_DRAFT = None, wait_tick_load: bool = True) -> bool:
    #     """
    #
    #     SYMBOL_NAME have to be in terminal! otherwise error
    #         [False]tick_last__update()=mt5.last_error()=(-4, 'Terminal: Not found')
    #     """
    #     _symbol = self.SYMBOL__get_active(_symbol)
    #     result = False
    #     while True:
    #         tick = mt5.symbol_info_tick(self.SYMBOL)
    #         result = tick != self.TICK_LAST
    #         if result:
    #             break
    #
    #         if not wait_tick_load:
    #             break
    #         time.sleep(1)
    #
    #     if result:
    #         print(f"update[{self.TICK_LAST=}]{_symbol}/{mt5.last_error()=}")
    #         # Tick(time=1665770358, bid=62.437, ask=63.312, last=0.0, volume=0, time_msc=1665770358179, flags=6, volume_real=0.0)
    #         self.TICK_LAST = tick
    #     return result
    #
    # HISTORY ---------------------------------------------------------------------------------------------------------
    # def bars__check_actual(
    #         self,
    #         _symbol: TYPING__SYMBOL_DRAFT = None,
    #         _tf: TYPING__TF = None
    # ) -> bool:
    #     _symbol = self.SYMBOL__get_active(_symbol)
    #     _tf_td = dt.timedelta(minutes=self.TF__get_active(_tf))
    #
    #     result = False
    #     last = self.TICK_LAST
    #     if last:
    #         last_dt = dt.datetime.fromtimestamp(last.time)
    #         result = (last_dt + _tf_td) >= dt.datetime.today()
    #     return result


# =====================================================================================================================
def _explore():
    obj = MT5()
    bar1 = obj.history__get(10)
    print(bar1)
    ObjectInfo(bar1).print()
    # print(bar1.dtype)
    # print(bar1.dtype.fields)
    # print()
    # print()

    # bars2 = obj.bars__get(2)
    # print(bars2)
    # ObjectInfo(bars2).print()


# =====================================================================================================================
if __name__ == "__main__":
    _explore()


# =====================================================================================================================
